from __future__ import unicode_literals

import datetime

from requests_oauthlib import OAuth2Session

from .constants import ACCESS_TOKEN_URL, AUTHORIZE_URL, MYOB_PARTNER_BASE_URL


class PartnerCredentials():
    """An object wrapping the 3-step OAuth2 process for Partner MYOB API access."""
    def __init__(
        self, consumer_key, consumer_secret, callback_uri,
        verified=False,
        userpass=None,
        oauth_token=None,
        refresh_token=None,
        oauth_expires_at=None,
        scope=None
    ):
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.callback_uri = callback_uri

        self.verified = verified
        self.userpass = userpass
        self.oauth_token = oauth_token
        self.refresh_token = refresh_token
        self.oauth_expires_at = oauth_expires_at

        self._oauth = OAuth2Session(consumer_key, redirect_uri=callback_uri)
        url, _ = self._oauth.authorization_url(MYOB_PARTNER_BASE_URL + AUTHORIZE_URL)
        self.url = url + '&scope=CompanyFile'

    @property
    def state(self):
        """ Get a representation of this credentials object from which it can be reconstructed. """
        return dict(
            (attr, getattr(self, attr))
            for attr in (
                'consumer_key', 'consumer_secret', 'callback_uri',
                'verified', 'userpass', 'oauth_token', 'refresh_token',
                'oauth_expires_at'
            )
            if getattr(self, attr) is not None
        )

    def expired(self, now=None):
        """ Determine whether the current access token has expired. """
        # Expiry might be unset if the user hasn't finished authenticating.
        if self.oauth_expires_at is None:
            return False

        # Allow a bit of time for clock differences and round trip times
        # to prevent false negatives. If users want the precise expiry,
        # they can use self.oauth_expires_at
        CONSERVATIVE_SECONDS = 30

        now = now or datetime.datetime.now()
        return self.oauth_expires_at <= (now + datetime.timedelta(seconds=CONSERVATIVE_SECONDS))

    def verify(self, code):
        """ Verify an OAuth session, retrieving an access token. """
        token = self._oauth.fetch_token(
            MYOB_PARTNER_BASE_URL + ACCESS_TOKEN_URL,
            code=code,
            client_secret=self.consumer_secret,
        )
        self.save_token(token)

    def refresh(self):
        """ Refresh an expired token. """
        token = self._oauth.refresh_token(
            MYOB_PARTNER_BASE_URL + ACCESS_TOKEN_URL,
            refresh_token=self.refresh_token,
            client_id=self.consumer_key,
            client_secret=self.consumer_secret,
        )
        self.save_token(token)

    def save_token(self, token):
        self.oauth_token = token.get('access_token')
        self.refresh_token = token.get('refresh_token')

        self.oauth_expires_at = datetime.datetime.fromtimestamp(token.get('expires_at'))
        self.verified = True



from .endpoints import ENDPOINTS
from .managers import Manager


class Myob:
    """An ORM-like interface to the MYOB API"""
    def __init__(self, credentials):
        from .credentials import PartnerCredentials
        if not isinstance(credentials, PartnerCredentials):
            raise TypeError(
                'Expected a Credentials instance, got %s.' % (
                    type(credentials).__name__,
                )
            )
        self.credentials = credentials

        for k, v in ENDPOINTS.items():
            setattr(self, v['plural'], Manager(k, credentials))

    def __repr__(self):
        return '%s:\n    %s' % (self.__class__.__name__, '\n    '.join(
            sorted(v['plural'] for v in ENDPOINTS.values())
        ))



ALL = 'ALL'
GET = 'GET'
POST = 'POST'
PUT = 'PUT'
DELETE = 'DELETE'

METHOD_ORDER = [ALL, GET, POST, PUT, DELETE]

ENDPOINTS = {
    '': {
        'plural': 'companyfiles',
        'methods': [
            (ALL, '', 'Return a list of company files.'),
            (GET, '[id]/', 'List endpoints available for a company file.'),
            (GET, 'Info/', 'Return API build information for each individual endpoint.'),
        ],
    },
    '[company_id]/Contact/': {
        'plural': 'contacts',
        'methods': [
            (ALL, '', 'Return all contact types for an AccountRight company file.'),
            (ALL, 'Customer/', 'Return all customer contacts for an AccountRight company file.'),
            (GET, 'Customer/[uid]/', 'Return selected customer contact.'),
            (PUT, 'Customer/[uid]/', 'Update selected customer contact.'),
            (POST, 'Customer/', 'Create new customer contact.'),
            (DELETE, 'Customer/[uid]/', 'Delete selected customer contact.'),
            (ALL, 'Supplier/', 'Return all supplier contacts for an AccountRight company file.'),
            (GET, 'Supplier/[uid]/', 'Return selected supplier contact.'),
            (PUT, 'Supplier/[uid]/', 'Update selected supplier contact.'),
            (POST, 'Supplier/', 'Create new supplier contact.'),
            (DELETE, 'Supplier/[uid]/', 'Delete selected supplier contact.'),
        ],
    },
    '[company_id]/Sale/Invoice/': {
        'plural': 'invoices',
        'methods': [
            (ALL, '', 'Return all sale invoice types for an AccountRight company file.'),
            (ALL, 'Item/', 'Return item type sale invoices for an AccountRight company file.'),
            (GET, 'Item/[uid]/', 'Return selected item type sale invoice.'),
            (PUT, 'Item/[uid]/', 'Update selected item type sale invoice.'),
            (POST, 'Item/', 'Create new item type sale invoice.'),
            (DELETE, 'Item/[uid]/', 'Delete selected item type sale invoice.'),
            (ALL, 'Service/', 'Return service type sale invoices for an AccountRight company file.'),
            (GET, 'Service/[uid]/', 'Return selected service type sale invoice.'),
            (PUT, 'Service/[uid]/', 'Update selected service type sale invoice.'),
            (POST, 'Service/', 'Create new service type sale invoice.'),
            (DELETE, 'Service/[uid]/', 'Delete selected service type sale invoice.'),
        ]
    },
    '[company_id]/GeneralLedger/': {
        'plural': 'general_ledger',
        'methods': [
            (ALL, 'TaxCode/', 'Return tax codes set up with an AccountRight company file.'),
            (GET, 'TaxCode/[uid]/', 'Return selected tax code.'),
            (PUT, 'TaxCode/[uid]/', 'Update selected tax codes.'),
            (POST, 'TaxCode/', 'Create new tax code.'),
            (DELETE, 'TaxCode/[uid]/', 'Delete selected tax code.'),
            (ALL, 'Account/', 'Return accounts set up with an AccountRight company file.'),
            (GET, 'Account/[uid]/', 'Return selected account.'),
            (PUT, 'Account/[uid]/', 'Update selected accounts.'),
            (POST, 'Account/', 'Create new account.'),
            (DELETE, 'Account/[uid]/', 'Delete selected account.'),
        ]
    },
    '[company_id]/Inventory/': {
        'plural': 'inventory',
        'methods': [
            (ALL, 'Item/', 'Return inventory items for an AccountRight company file.'),
            (GET, 'Item/[uid]/', 'Return selected inventory item.'),
            (PUT, 'Item/[uid]/', 'Update selected inventory items.'),
            (POST, 'Item/', 'Create new inventory item.'),
            (DELETE, 'Item/[uid]/', 'Delete selected inventory item.'),
        ]
    },
    '[company_id]/Purchase/Order/': {
        'plural': 'purchase_orders',
        'methods': [
            (ALL, '', 'Return all purchase order types for an AccountRight company file.'),
            (ALL, 'Item/', 'Return item type purchase orders for an AccountRight company file.'),
            (GET, 'Item/[uid]/', 'Return selected item type purchase order.'),
            (PUT, 'Item/[uid]/', 'Update selected item type purchase order.'),
            (POST, 'Item/', 'Create new item type purchase order.'),
            (DELETE, 'Item/[uid]/', 'Delete selected item type purchase order.'),
        ]
    },
    '[company_id]/Purchase/Bill/': {
        'plural': 'purchase_bills',
        'methods': [
            (ALL, '', 'Return all purchase bill types for an AccountRight company file.'),
            (ALL, 'Item/', 'Return item type purchase bills for an AccountRight company file.'),
            (GET, 'Item/[uid]/', 'Return selected item type purchase bill.'),
            (PUT, 'Item/[uid]/', 'Update selected item type purchase bill.'),
            (POST, 'Item/', 'Create new item type purchase bill.'),
            (DELETE, 'Item/[uid]/', 'Delete selected item type purchase bill.'),
            (ALL, 'Service/', 'Return service type purchase bills for an AccountRight company file.'),
            (GET, 'Service/[uid]/', 'Return selected service type purchase bill.'),
            (PUT, 'Service/[uid]/', 'Update selected service type purchase bill.'),
            (POST, 'Service/', 'Create new service type purchase bill.'),
            (DELETE, 'Service/[uid]/', 'Delete selected service type purchase bill.'),
        ]
    },
}
